'''Script introducing machine-learning through sklearn (Edward Hirst)'''
#Import libraries
import numpy as np            #...useful data manipulation with arrays
from ast import literal_eval  #...used to interpret txt files of arrays
#Import used sklearn functions
from sklearn.neural_network import MLPRegressor, MLPClassifier
from sklearn.linear_model import LogisticRegression as LR
from sklearn.svm import SVC
from sklearn.ensemble import RandomForestClassifier as RF
from sklearn.model_selection import train_test_split  
from sklearn.metrics import mean_squared_error as MSE
from sklearn.metrics import confusion_matrix as CM

#%% #Import data (provided by the KS database) ...'#%%' denotes a new cell in spyder
#Import weight data
with open('./CY.txt','r') as f0:
    Weights = np.array(literal_eval(f0.read()))
    
#Import Hodge numbers file
with open('./Hodge.txt','r') as f1:
    Hodge = np.array(literal_eval(f1.read()))
    
del(f0,f1)

#%% #Regression: ML Hodge numbers from weights with NNs (loss MSE preset)
#Allocate input (X), and output (Y) data
X = Weights
Y = Hodge    #...variants: [:,0], [:,1], 2*(Hodge[:,0]-Hodge[:,1])

#Perform train/test splitting
X_train, X_test, Y_train, Y_test = train_test_split(X,Y,test_size=0.2,random_state=3)

#Define & train NN Regressor on the train data
nn_reg = MLPRegressor(hidden_layer_sizes=(32,64,32),activation='relu',solver='adam',batch_size=32,max_iter=20,verbose=True)
nn_reg.fit(X_train, Y_train) 

#Calculate predictions with NN Regressor on the test data
Y_pred = nn_reg.predict(X_test)
mse_loss = MSE(Y_test,Y_pred)
R_sq = nn_reg.score(X_test,Y_test)

#Output evaluated learning measure
print('\nNN Regressor:\nMSE: '+str(mse_loss)+'\nR^2 score: '+str(R_sq)+'\n')

#%% #Classification: ML Euler number positivity from weights with NNs (loss log-loss preset)
X = Weights
Y = np.array([])
for h in Hodge:
    if 2*(h[0]-h[1]) > 0:   Y = np.append(Y,[1])
    else:                   Y = np.append(Y,[0])

#Perform train/test splitting
X_train, X_test, Y_train, Y_test = train_test_split(X,Y,test_size=0.2,random_state=22)

#Intialise and train NN classifier on the train data
nn_clf = MLPClassifier((32,64,32),activation='relu',solver='adam',batch_size=32,max_iter=20,verbose=True)
nn_clf.fit(X_train, Y_train) 

#Calculate predictions with NN classifier on the test data
Y_pred = nn_clf.predict(X_test)
clf_CM = CM(Y_test,Y_pred,normalize='all')
Accuracy = np.sum(Y_pred == Y_test)/len(Y_test)
  
print('\nNN Classifier:\nAccuracy: '+str(Accuracy)+'\nConfusion-Matrix:\n'+str(np.matrix(clf_CM))+'\n')

#%% #Classification: ML Euler number positivity from weights - Logistic Regressor
X = Weights
Y = np.array([])
for h in Hodge:
    if 2*(h[0]-h[1]) > 0:   Y = np.append(Y,[1])
    else:                   Y = np.append(Y,[0])

#Perform train/test splitting
X_train, X_test, Y_train, Y_test = train_test_split(X,Y,test_size=0.2,random_state=14)

#Intialise and train Logistic Regressor
lr_clf = LR(tol=1,C=100,random_state=1) #...tol is tolerance for stopping training, C denotes inverse weighting of regularisation (high C => low regularisation => don't try to minimise weight values or use less weights)
lr_clf.fit(X_train, Y_train) 

#Calculate predictions with Logistic Regressor and evaluate testing
Y_pred = lr_clf.predict(X_test)
clf_CM = CM(Y_test,Y_pred,normalize='all')
Accuracy = np.sum(Y_pred == Y_test)/len(Y_test)
  
print('LR:\nAccuracy: '+str(Accuracy)+'\nConfusion-Matrix:\n'+str(np.matrix(clf_CM)))

#%% #Classification: ML Euler number positivity from weights - Support Vector Machine
X = Weights
Y = np.array([])
for h in Hodge:
    if 2*(h[0]-h[1]) > 0:   Y = np.append(Y,[1])
    else:                   Y = np.append(Y,[0])

#Perform train/test splitting
X_train, X_test, Y_train, Y_test = train_test_split(X,Y,test_size=0.2,random_state=53)

#Intialise and train Support Vector Machine
svm_clf = SVC(C=1,kernel='linear',random_state=1) #...SVC --> clf, SVR --> regressor
svm_clf.fit(X_train, Y_train) 

#Calculate predictions with Support Vector Machine and evaluate testing
Y_pred = svm_clf.predict(X_test)
clf_CM = CM(Y_test,Y_pred,normalize='all')
Accuracy = np.sum(Y_pred == Y_test)/len(Y_test)
  
print('SVM:\nAccuracy: '+str(Accuracy)+'\nConfusion-Matrix:\n'+str(np.matrix(clf_CM)))  

#%% #Classification: ML Euler number positivity from weights - Random Forest
X = Weights
Y = np.array([])
for h in Hodge:
    if 2*(h[0]-h[1]) > 0:   Y = np.append(Y,[1])
    else:                   Y = np.append(Y,[0])

#Perform train/test splitting
X_train, X_test, Y_train, Y_test = train_test_split(X,Y,test_size=0.2,random_state=99)

#Intialise and train Support Vector Machine
rf_clf = RF(n_estimators=100,max_depth=2,random_state=44)
rf_clf.fit(X_train, Y_train) 

#Calculate predictions with Support Vector Machine and evaluate testing
Y_pred = rf_clf.predict(X_test)
clf_CM = CM(Y_test,Y_pred,normalize='all')
Accuracy = np.sum(Y_pred == Y_test)/len(Y_test)
  
print('RF:\nAccuracy: '+str(Accuracy)+'\nConfusion-Matrix:\n'+str(np.matrix(clf_CM)))  


########################################################################################
'''
References:
This script uses data from the KS database:
@misc{kreuzer2000complete,
      title={Complete classification of reflexive polyhedra in four dimensions}, 
      author={Maximilian Kreuzer and Harald Skarke},
      year={2000},
      eprint={hep-th/0002240},
      archivePrefix={arXiv},
      primaryClass={hep-th}
}
...and then uses code from the project ML-ing CY3 hypersurfaces:
@article{Berman:2021mcw,
    author = "Berman, David S. and He, Yang-Hui and Hirst, Edward",
    title = "{Machine Learning Calabi-Yau Hypersurfaces}",
    eprint = "2112.06350",
    archivePrefix = "arXiv",
    primaryClass = "hep-th",
    reportNumber = "QMUL-PH-21-55, LIMS-2021-017",
    month = "12",
    year = "2021"
}
If you end up using some of this code please drop us a citation :) 
'''